///////////////////////////////////////////////////////////////////////////////
// Copyright (C) 2002 Jason Baldridge and Gann Bierner
// 
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
// 
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//////////////////////////////////////////////////////////////////////////////

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import opennlp.maxent.Evalable;
import opennlp.maxent.EventCollector;
import opennlp.maxent.EventStream;
import opennlp.maxent.GISModel;
import opennlp.maxent.MaxentModel;
import opennlp.maxent.TrainEval;
import opennlp.maxent.PlainTextByLineDataStream;
import opennlp.maxent.TwoPassDataIndexer;
import opennlp.maxent.EventCollectorAsStream;
import opennlp.maxent.io.SuffixSensitiveGISModelReader;
import opennlp.maxent.io.SuffixSensitiveGISModelWriter;
import opennlp.tools.util.Sequence;

/**
 * A part-of-speech tagger that uses maximum entropy.  Trys to predict whether
 * words are nouns, verbs, or any of 70 other POS tags depending on their
 * surrounding context.
 *
 * @author      Gann Bierner
 * @version $Revision: 1.3 $, $Date: 2007/01/29 16:37:24 $
 */
public abstract class AbstractSuperTaggerME implements Evalable, SuperTagger {

  /**
   * The maximum entropy model to use to evaluate contexts.
   */
    protected MaxentModel _posModel;

  /**
   * The feature context generator.
   */
    protected AbstractContextGenerator _contextGen;

  /**
   * Says whether a filter should be used to check whether a tag assignment
   * is to a word outside of a closed class.
   */
    protected boolean _useClosedClassTagsFilter = false;
  
    protected static int DEFAULT_BEAM_SIZE = 3;
    protected static int NUM_SOLUTIONS =  1;

  /** The size of the beam to be used in determining the best sequence of pos tags.*/
    protected int size;

    protected Sequence[] bestSequences;
  
  /** The search object used for search multiple sequences of tags. */
    protected BeamSearch beam;
  
    protected AbstractSuperTaggerME(MaxentModel mod, AbstractContextGenerator cg) {
	this(DEFAULT_BEAM_SIZE, mod, cg);
    }

    protected AbstractSuperTaggerME(int beamSize, MaxentModel mod, AbstractContextGenerator cg) {
	size = beamSize;
	_posModel = mod;
	_contextGen = cg;
	beam = new STBeamSearch(size, cg, mod);
    }

    public String getNegativeOutcome() {
	return "";
    }
  
  /**
   * Returns the number of different tags predicted by this model.
   * @return the number of different tags predicted by this model.
   */
    public int getNumTags() {
	return _posModel.getNumOutcomes();
    }

    public EventCollector getEventCollector(Reader r) {
	return new STEventCollector(r, _contextGen);
    }

    public Pair splitString(String s) {
	int split = s.indexOf("/");
	return new Pair(s.substring(0,split), s.substring(split+1));
    }

    public List[] supertag(List wordtags) {
	return supertag(NUM_SOLUTIONS, wordtags);
    }
	
    public List[] supertag(int NumSolutions, List wordtags) {
	bestSequences = beam.bestSequences(NumSolutions, wordtags.toArray(), null);
	List[] outcomes = new ArrayList[NumSolutions];
	for (int i=0; i<NumSolutions && i<bestSequences.length; i++) {
	    outcomes[i] = bestSequences[i].getOutcomes();
	}
	return outcomes;
    }

    public String[] supertag(String[] wordtags) {
	return supertag(NUM_SOLUTIONS, wordtags);
    }

    public String[] supertag(int NumSolutions, String[] wordtags) {
	List wt = new ArrayList(wordtags.length);
	for (int i=0; i<wordtags.length; i++) {
	    wt.add(splitString(wordtags[i]));
	}
	List[] t = supertag(NumSolutions, wt);
	String[] r = new String[wordtags.length];
	for (int i=0; i<wordtags.length; i++) {
	    StringBuffer sb = new StringBuffer();
	    sb.append(t[0].get(i));
	    for(int j=1; j<bestSequences.length; j++) {
		sb.append("-" + t[j].get(i));
	    }
	    r[i]=sb.toString();
	}
	return r;
    }

    public String supertag(String taggedsentence) {
	return supertag(NUM_SOLUTIONS, taggedsentence);
    }

    public String supertag(int NumSolutions, String taggedsentence) {
	ArrayList wordtags = new ArrayList();
	StringTokenizer st = new StringTokenizer(taggedsentence);
	while(st.hasMoreTokens()) {
	    String s = (String) st.nextToken();
	    wordtags.add(splitString(s));
	}
	List[] supertags = supertag(NumSolutions, wordtags);
	StringBuffer sb = new StringBuffer();
	for(int i=0; i<wordtags.size(); i++) {
	    sb.append(wordtags.get(i).toString() + "/");
	    sb.append(supertags[0].get(i));
	    for(int j=1; j<bestSequences.length; j++) {
		sb.append("-" + supertags[j].get(i));
	    }
	    sb.append(" ");
	}
	return sb.toString().trim();
    }


    public void probs(double[] probs) {
	bestSequences[0].getProbs(probs);
    }

    public double[] probs() {
	return bestSequences[0].getProbs();
    }


    public void localEval(MaxentModel posModel, Reader r, Evalable e, boolean verbose) {

	_posModel = posModel;
	float total = 0, correct = 0, sentences = 0, sentsCorrect = 0;
	BufferedReader br = new BufferedReader(r);
	String line;
	try {
	    while ((line = br.readLine()) != null) {
		sentences++;
		Pair p = STEventCollector.convertAnnotatedString(line);
		List wordtags = (List) p.a;
		List outcomes = (List) p.b;
		List stags = beam.bestSequences(NUM_SOLUTIONS, wordtags.toArray(), null)[0].getOutcomes();
		
        int c = 0;
        boolean sentOk = true;
        for (Iterator t = stags.iterator(); t.hasNext(); c++) {
	    total++;
	    String tag = (String) t.next();
	    if (stags.equals(outcomes.get(c)))
		correct++;
	    else
		sentOk = false;
        }
        if (sentOk)
	    sentsCorrect++;
	    }
	}
	catch (IOException E) {
	    E.printStackTrace();
	}

	System.out.println("Accuracy         : " + correct / total);
	System.out.println("Sentence Accuracy: " + sentsCorrect / sentences);

  }

    protected class STBeamSearch extends BeamSearch {
	
	public STBeamSearch(int size, AbstractContextGenerator cg, MaxentModel model) {
	    super(size, cg, model);
	}
    
	public STBeamSearch(int size, AbstractContextGenerator cg, MaxentModel model, int cacheSize) {
	    super(size, cg, model, cacheSize);
	}

 
	protected boolean validSequence(int i, Object[] inputSequence, String[] outcomesSequence, String outcome) {
	    /* reject `unknown' supertags */
	    return !outcome.equals("0");
	}
      
	protected boolean validSequence(int i, List inputSequence, Sequence outcomesSequence, String outcome) {
	    /* reject `unknown' supertags */
	    return !outcome.equals("0");
	}
    }
  
    public String[] getOrderedTags(List words, List tags, List supertags, int index) {
	return getOrderedTags(words,tags,supertags,index,null);
    }
  
    public String[] getOrderedTags(List words, List tags, List supertags, int index,double[] tprobs) {
	double[] probs = _posModel.eval(_contextGen.getContext(index,words.toArray(),tags.toArray(), (String[]) supertags.toArray(new String[tags.size()]),null));
	String[] orderedTags = new String[probs.length];
	for (int i = 0; i < probs.length; i++) {
	    int max = 0;
	    for (int ti = 1; ti < probs.length; ti++) {
		if (probs[ti] > probs[max]) {
		    max = ti;
		}
	    }
	    orderedTags[i] = _posModel.getOutcome(max);
	    if (tprobs != null){
		tprobs[i]=probs[max];
	    }
	    probs[max] = 0;
	}
	return (orderedTags);
    }

    public static GISModel train(EventStream es, int iterations, int cut) throws IOException {
	return opennlp.maxent.GIS.trainModel(iterations, new TwoPassDataIndexer(es, cut));
    }
    
}
